//ModbusRTU.cpp

#include <Arduino.h>
#include "config.h"
#include "ModbusRTU.h"
#include "maindat.h"

DAT_MRTU  dmrtu;


//инициализация
int16_t  sMRTU_INIT(void)
{
  memset(&dmrtu.rxbuf[0], 0, _szbuf);
  memset(&dmrtu.txbuf[0], 0, _szbuf);
  dmrtu.adrx = 0;
  dmrtu.adtx = 0;
  return 0;
}

//контрольная сумма
uint16_t sMRTU_CRC(uint8_t* data, uint16_t len)
{
  uint16_t res_CRC = 0xFFFF;
  uint16_t count = 0;
  uint8_t count_crc;
  uint8_t dt;
while(count < len)
{
  //
  count_crc = 0;
  dt = (uint8_t)(data[count]);
  res_CRC ^= (uint16_t)(dt);
  //
    while(count_crc < 8)
    {
      if((res_CRC & 0x0001) < 1)
      {
        res_CRC = (res_CRC >> 1) & 0x7FFF;
      }
      else
      {
        res_CRC = (res_CRC >> 1) & 0x7FFF;
        res_CRC ^= 0xA001;
      };
      count_crc++;
    };
    count++;
  }
  return (res_CRC);
}

//разбор команд
int16_t  sMRTU_CMD(uint8_t * dbuff, uint8_t rxsz, uint8_t * tbuff)
{
  uint16_t  sreg  = 0;  //стартовый регистр
  uint16_t  cntreg  = 0;  //количество регистров
  uint16_t  datreg = 0;
  uint16_t  crc; //подсчитанная контрольная сумма
  uint16_t  incrc;          //подсчитанная контрольная сумма
  uint16_t  csize;
  uint8_t  workbuff[_szbuf];
  uint8_t  cmd = (*(uint8_t*)(dbuff + _ADDR_MCMD)); 
  switch(cmd)
  {
    case 0x03: //Read Holding Registers
    case 0x04:  //Read Input Registers
      csize  = 6;
      sreg  |= (* (uint16_t*)(dbuff + _ADDR_03_START_REG_H)) << 8;
      sreg  |= (*(uint16_t*)(dbuff + _ADDR_03_START_REG_H + 1));
      cntreg  |= (* (uint16_t*)(dbuff + _ADDR_03_CNT_REG_H)) << 8;
      cntreg  |= (*(uint16_t*)(dbuff + _ADDR_03_CNT_REG_H + 1));
      incrc  = (*(uint16_t*)(dbuff + _ADDR_03_INCRC));
      crc = sMRTU_CRC(dbuff, csize);
      if(crc == incrc)
      {
        crc = 0;
        //
        
        //
        return 0;
      }
      break;
    /*
    *****************************************************
    Запись в один регистр - ответ повторяет приём
    *****************************************************
    */
    case 0x06:  //Preset Single Register
      csize  = 6;
      sreg  |=(uint16_t) (* (uint8_t*)(dbuff + _ADDR_06_ADDR_REG_H)) << 8;
      sreg  |= (uint16_t) (*(uint8_t*)(dbuff + _ADDR_06_ADDR_REG_H + 1));
      datreg  |= (uint16_t) (* (uint8_t*)(dbuff +  _ADDR_06_DAT_H)) << 8;
      datreg  |= (uint16_t) (*(uint8_t*)(dbuff +  _ADDR_06_DAT_H + 1));
      incrc  = (*(uint16_t*)(dbuff + _ADDR_06_INCRC));
      crc = sMRTU_CRC(dbuff, csize);
      if(crc == incrc)
      {
        /*
        ****************************
        проверка границ данных
        ****************************
        */
        //ComMotion
        if((sreg >= _ComMotion_RMS) && (sreg < (_ComMotion_RMS + _ComMotion_SzRM)))
        {
           dmotor[sreg - _ComMotion_RMS] = (int16_t)(datreg);
           memcpy(tbuff, dbuff, 8);
           Serial.write(tbuff, 8);
           Serial.flush();
        }
        //Multiservo
        else if((sreg >= _Multiservo_RSS1) && (sreg < (_Multiservo_RSS1 + _Multiservo_SzRS)))
        {
          dservo[sreg - _Multiservo_RSS1] = (int16_t)(datreg);
          memcpy(tbuff, dbuff, 8);
          Serial.write(tbuff, 8);
          Serial.flush();
        }
        else
          Serial.println("Addr data error");
        //
        return 100;
      }
      break;
    default:  //неизвесная команда
      Serial.println("Umcnown command");
      Serial.flush();
      return _ILLEGAL_FUNCTION;
  }
  return 1;
}

//приём команды и обработка
int16_t  MRTU_rWORK(void)
{
  int rcvsz = Serial.available();
  uint8_t  resb;
  int rcnt = 0, rccnt = 0;
  uint8_t wcnt = 0;
  //нет пакета для обработки
  if(rcvsz < 8)
    return 0;
  //Serial.println("Working");
  //Serial.flush();
  //копирование данных в буфер приёма
/*  while(rcnt < rcvsz)
  {
    resb = Serial.read();
    if(rccnt < _szbuf)
    {
      dmrtu.rxbuf[rccnt] = resb;
      wcnt++;
    }
    rcnt++;
    rccnt++;
  }
*/
  if(rcvsz < _szbuf)
    wcnt = rcvsz;
  else
    wcnt = _szbuf;
  Serial.readBytes((char*)(&dmrtu.rxbuf[0]), wcnt);
  while(rcnt < rcvsz)
  {
    resb = Serial.read();
    rcnt++;
  }
  if(dmrtu.rxbuf[0] != _MRTU_Addr)
  {
    Serial.println("No main device");
    Serial.flush();
    return 0;
  }
  //обработка принятых данных
  return sMRTU_CMD(&dmrtu.rxbuf[0], wcnt, &dmrtu.txbuf[0]);
}

//передача ответа
int16_t  MRTU_sWORK(void)
{
  return 0;
}

